// WL_AGENT.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                                LOCAL CONSTANTS

=============================================================================
*/

#define MAXMOUSETURN    10


#define MOVESCALE       150l
#define BACKMOVESCALE   100l
#define ANGLESCALE      20

#define PlayItemSound(s)    PlaySoundGlobal(s,chan_item,0)

/*
=============================================================================

                                GLOBAL VARIABLES

=============================================================================
*/

short maxammo[NUMAMMOTYPES] =
{
    200,    // ammo_bullet
    100,    // ammo_rifle
    100,    // ammo_stg44
    200,    // ammo_flame
    400,    // ammo_chaingun
     20,    // ammo_rocket
    200,    // ammo_flak
};

//
// player state info
//
int32_t         thrustspeed;

word            plux,pluy;          // player coordinates scaled to unsigned

short           anglefrac;

objtype        *LastAttacker;
damage_t       KillDamageType;

/*
=============================================================================

                                                 LOCAL VARIABLES

=============================================================================
*/


void    T_Player (objtype *ob);
void    T_Attack (objtype *ob);

statetype   s_player = {false,0,0,(statefunc) T_Player,NULL,NULL};
statetype   s_attack = {false,0,0,(statefunc) T_Attack,NULL,NULL};

/*
WSJ: Attack Reference

   -3 - end of firing sequence if no ammo left (bolt sound)
   -2 - end of firing sequence if no ammo left
   -1 - end of firing sequence
    1 - gun attack
    2 - knife attack
    3 - mp40 refire
    4 - chaingun refire
    5 - gun attack (no sound)
    6 - chaingun refire (no sound)
*/

struct atkinf
{
    int8_t    tics,attack,frame;
} attackinfo[NUMWEAPONS][15] =
{
    { {6,0,1},{6,2,2},{6,0,3},{6,-1,4} },
    { {6,0,1},{6,1,2},{6,0,3},{6,-1,4} },
    { {6,0,1},{6,1,2},{10,-2,3},{6,1,4},{10,3,5},{6,-1,6} },
    { {6,0,1},{3,1,2},{3,-2,3},{3,1,2},{3,-2,3},{3,1,2},{3,-2,3},{3,1,2},{3,-2,3},{3,1,2},{14,0,3},{3,3,3},{6,-1,4} },
    { {6,0,1},{3,1,2},{3,1,3},{3,1,2},{3,1,3},{3,1,2},{3,1,3},{3,1,2},{3,1,3},{3,1,2},{3,1,3},{14,0,4},{3,3,4},{6,-1,4} },
    { {6,0,1},{6,1,2},{10,0,3},{10,-3,4},{10,0,5},{6,0,6},{6,0,7},{6,0,8},{10,0,9},{10,0,10},{6,-1,11} },
    { {6,0,0},{6,1,0},{10,0,0},{10,-3,0},{10,0,0},{6,0,0},{6,0,0},{6,0,0},{10,0,0},{10,0,0},{6,-1,0} },
    { {5,0,1},{5,1,2},{5,3,3},{5,-1,4} },
    { {5,0,1},{5,1,2},{5,4,3},{5,-1,4} },
    { {5,0,1},{5,1,2},{5,3,3},{5,-1,4} },
    { {4,0,1},{4,1,2},{4,5,3},{4,5,2},{4,5,3},{4,5,2},{4,5,3},{4,5,2},{4,6,3},{6,-1,4} },
    { {5,0,1},{5,1,2},{5,5,3},{5,1,2},{5,6,3},{6,-1,4} },
    { {4,0,1},{4,1,2},{4,5,3},{4,5,2},{4,5,3},{4,1,2},{4,5,3},{4,5,2},{4,6,3},{6,-1,4} },
    { {6,0,1},{6,1,2},{6,0,3},{64,0,4},{6,-1,5} },
    { {7,0,1},{7,1,2},{7,4,3},{7,-1,4} },
    { {5,0,1},{5,1,2},{5,5,3},{5,1,2},{5,6,3},{6,-1,4} },
};

//===========================================================================

//----------

void Attack (void);
void Use (void);
void Search (objtype *ob);
void SelectWeapon (void);
void SelectItem (void);

//----------

boolean TryMove (objtype *ob);
void T_Player (objtype *ob);

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove);

/*
=============================================================================

                                CONTROL STUFF

=============================================================================
*/

/*
======================
=
= CheckWeaponChange
=
= Number keys change weapons
=
======================
*/

short weaponkey[NUMWEAPONS] =
{
    bt_readyweapon1,     // wp_knife
    bt_readyweapon2,     // wp_pistol
    bt_readyweapon2,     // wp_dualpistol
    bt_readyweapon2,     // wp_mauser
    bt_readyweapon2,     // wp_dualmauser
    bt_readyweapon3,     // wp_kar98k
    bt_readyweapon3,     // wp_sniper
    bt_readyweapon4,     // wp_mp40
    bt_readyweapon4,     // wp_dualmp40
    bt_readyweapon5,     // wp_stg44
    bt_readyweapon6,     // wp_flamethrower
    bt_readyweapon7,     // wp_mg42
    bt_readyweapon7,     // wp_chaingun
    bt_readyweapon8,     // wp_panzerschreck
    bt_readyweapon9      // wp_flakgun
};

// WSJ: check the type of ammo used by weapon
ammo_t CheckAmmoType (int weapon)
{
    switch (weapon)
    {
     case wp_pistol:
     case wp_dualpistol:
     case wp_mauser:
     case wp_dualmauser:
     case wp_mp40:
     case wp_dualmp40:
         return ammo_bullet;
     case wp_kar98k:
     case wp_sniper:
         return ammo_rifle;
     case wp_stg44:
         return ammo_stg44;
     case wp_flamethrower:
         return ammo_flame;
     case wp_mg42:
     case wp_chaingun:
         return ammo_chaingun;
     case wp_panzerschreck:
         return ammo_rocket;
     case wp_flakgun:
         return ammo_flak;
    }

    return ammo_none;
}

boolean HasAmmo (int weapon)
{
    if (CheckAmmoType(weapon) == ammo_none) return true;
    else if (gamestate.ammo[CheckAmmoType(weapon)] > 0) return true;
    return false;
}

boolean WeaponAvailable (int weapon)
{
    return (gamestate.hasweapon[weapon] && HasAmmo(weapon));
}

weapontype BestAvailableWeapon (boolean skipdangerweapon)
{
    int bestweapon = 0;

    for (int i = 0; i <= gamestate.bestweapon; i++)
    {
        switch (i)
        {
            case wp_sniper:
            case wp_panzerschreck:
            case wp_flakgun:
                if (skipdangerweapon)
                    continue;
            default:
                if (WeaponAvailable(i))
                    bestweapon = i;
        }
    }

    return (weapontype)bestweapon;
}

weapontype CheckWeaponChoice (short wpkey)
{
    weapontype firstchoice = wp_noweapon;
    weapontype lastchosen = gamestate.lastchosenweapon[wpkey];

    if (wpkey != weaponkey[gamestate.weapon] && lastchosen > 0
    && WeaponAvailable(lastchosen))
        return lastchosen;

    for (int i = 0; i < NUMWEAPONS; i++)
    {
        if (!WeaponAvailable(i) || weaponkey[i] != wpkey)
            continue;

        if (firstchoice == -1) firstchoice = (weapontype)i;

        if (wpkey == weaponkey[gamestate.weapon] && i > gamestate.weapon)
            return (weapontype)i;
    }

    return firstchoice;
}

void CheckWeaponChange (void)
{
    int newWeapon = -1;
    weapontype bestweapon = BestAvailableWeapon(false);

    if(gamestate.weapon >= wp_mg42_mounted)    // WSJ: can't switch while using a mounted weapon
        return;

#ifdef _arch_dreamcast
    int joyx, joyy;
    IN_GetJoyFineDelta (&joyx, &joyy);
    if(joyx < -64)
        buttonstate[bt_prevweapon] = true;
    else if(joyx > 64)
        buttonstate[bt_nextweapon] = true;
#endif

    if(buttonstate[bt_nextweapon] && !buttonheld[bt_nextweapon])
    {
        if (gamestate.weapon+1 > bestweapon) newWeapon = 0;
        else
        {
            for(int i = gamestate.weapon+1; i <= bestweapon; i++)
            {
                if(WeaponAvailable(i))
                {
                    newWeapon = i;
                    break;
                }
            }
        }
    }
    else if(buttonstate[bt_prevweapon] && !buttonheld[bt_prevweapon])
    {
        if(gamestate.weapon-1 < 0) newWeapon = bestweapon;
        else
        {
            for(int i = gamestate.weapon-1; i >= 0; i--)
            {
                if(WeaponAvailable(i))
                {
                    newWeapon = i;
                    break;
                }
            }
        }
    }
    else
    {
        for(int i = bt_readyweapon1; i <= bt_readyweapon10; i++)
        {
            if (buttonstate[i] && !buttonheld[i])
            {
                newWeapon = CheckWeaponChoice(i);
                break;
            }
        }
    }

    if(newWeapon != -1)
    {
        gamestate.weapon = gamestate.chosenweapon = (weapontype) newWeapon;
        if (newWeapon != wp_sniper)
            gamestate.lastchosenweapon[weaponkey[newWeapon]] = (weapontype) newWeapon;
        DrawAmmo();
    }
}


/*
=======================
=
= ControlMovement
=
= Takes controlx,controly, and buttonstate[bt_strafe]
=
= Changes the player's angle and position
=
= There is an angle hack because when going 70 fps, the roundoff becomes
= significant
=
=======================
*/

void ControlMovement (objtype *ob)
{
    int32_t oldx,oldy,strafemove;
    int     angle;
    int     angleunits;

    if(buttonstate[bt_run])
        strafemove = RUNMOVE;
    else
        strafemove = BASEMOVE;

    if (SNIPERMODE)
        strafemove >>= 2;

    thrustspeed = 0;

    oldx = player->x;
    oldy = player->y;

    if(buttonstate[bt_strafeleft])
    {
        angle = ob->angle + ANGLES/4;
        if(angle >= ANGLES)
            angle -= ANGLES;
        Thrust(angle, strafemove * MOVESCALE * tics);
    }

    if(buttonstate[bt_straferight])
    {
        angle = ob->angle - ANGLES/4;
        if(angle < 0)
            angle += ANGLES;
        Thrust(angle, strafemove * MOVESCALE * tics);
    }

    //
    // side to side move
    //
    if (buttonstate[bt_strafe])
    {
        //
        // strafing
        //
        //
        if (controlx > 0)
        {
            angle = ob->angle - ANGLES/4;
            if (angle < 0)
                angle += ANGLES;
            Thrust (angle,controlx*MOVESCALE);      // move to left
        }
        else if (controlx < 0)
        {
            angle = ob->angle + ANGLES/4;
            if (angle >= ANGLES)
                angle -= ANGLES;
            Thrust (angle,-controlx*MOVESCALE);     // move to right
        }
    }
    else
    {
        //
        // not strafing
        //
        anglefrac += controlx;
        angleunits = anglefrac/ANGLESCALE;
        anglefrac -= angleunits*ANGLESCALE;
        ob->angle -= angleunits;

        if (ob->angle >= ANGLES)
            ob->angle -= ANGLES;
        if (ob->angle < 0)
            ob->angle += ANGLES;

    }

    //
    // forward/backwards move
    //
    if (controly < 0)
    {
        Thrust (ob->angle,-controly*MOVESCALE); // move forwards
    }
    else if (controly > 0)
    {
        angle = ob->angle + ANGLES/2;
        if (angle >= ANGLES)
            angle -= ANGLES;
        Thrust (angle,controly*BACKMOVESCALE);          // move backwards
    }

    if (gamestate.victoryflag)              // watching the BJ actor
        return;
}

/*
=============================================================================

                            STATUS WINDOW STUFF

=============================================================================
*/


/*
==================
=
= StatusDrawPic
=
==================
*/

void StatusDrawPic (unsigned x, unsigned y, unsigned picnum)
{
    LatchDrawPicScaledCoord ((screenWidth-scaleFactor*320)/16 + scaleFactor*x,
        screenHeight-scaleFactor*(STATUSLINES-y),picnum);
}

void StatusDrawFace(unsigned picnum)
{
    StatusDrawPic(17, 4, picnum);
    StatusDrawLCD(picnum);
}


/*
==================
=
= DrawFace
=
==================
*/

void DrawFace (void)
{
    if(viewsize == 21 && ingame) return;
    if (gamestate.health)
    {
        if (gamestate.attackframe || SNIPERMODE)
            gamestate.faceframe = 0;
#ifdef SPEAR
        if (godmode)
            StatusDrawFace(GODMODEFACE1PIC+gamestate.faceframe);
        else
#endif
            StatusDrawFace(FACE1APIC+3*((100-gamestate.health)/16)+gamestate.faceframe);
    }
    else
    {
#ifndef SPEAR
        if (KillDamageType == dam_mutagen)
            StatusDrawFace(MUTANTBJPIC);
        else
#endif
            StatusDrawFace(FACE8APIC);
    }
}

/*
===============
=
= UpdateFace
=
= Calls draw face if time to change
=
===============
*/

int facecount = 0;
int facetimes = 0;

void UpdateFace (void)
{
    // don't make demo depend on sound playback
    if(demoplayback || demorecord)
    {
        if(facetimes > 0)
        {
            facetimes--;
            return;
        }
    }

    facecount += tics;
    if (facecount > (Random(100)+20))
    {
        gamestate.faceframe = Random(3);
        if (gamestate.faceframe>2)
            gamestate.faceframe = 1;

        facecount = 0;
        DrawFace ();
    }
}



/*
===============
=
= LatchNumber
=
= right justifies and pads with blanks
=
===============
*/

static void LatchNumber (int x, int y, unsigned width, int32_t number)
{
    unsigned length,c;
    char    str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        StatusDrawPic (x,y,N_BLANKPIC);
        x++;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        if (number < 0)
            StatusDrawPic (x,y,N_BLANKPIC);
        else
            StatusDrawPic (x,y,str[c]-'0'+ N_0PIC);
        x++;
        c++;
    }
}


/*
===============
=
= DrawHealth
=
===============
*/

void DrawHealth (void)
{
    if(viewsize == 21 && ingame) return;
    LatchNumber (21,16,3,gamestate.health);
    LatchNumber (33,16,3,gamestate.armor);
}


/*
===============
=
= TakeDamage
=
===============
*/

void TakeDamage (int points, objtype *attacker, damage_t damagetype, boolean ignorearmor, boolean useflash)
{
    LastAttacker = attacker;

    if (points < 1 || gamestate.victoryflag || godmode)
        return;

    if (gamestate.difficulty==gd_baby)
    {
        points >>= 1;
        if (!points) points = 1;        // WSJ: don't make player entirely invulnerable
    }

    if (!ignorearmor)
    {
        if (gamestate.weapon >= wp_mg42_mounted)
        {
            points >>= 1;               // WSJ: take less damage while using mounted weapon,
            if (!points) points = 1;    // but still don't make the player invulnerable
        }

        if (gamestate.armor)
        {
            int reducedpoints = points/((100+gamestate.armor)/50);
            gamestate.armor -= (points>>1)+1;
            points = reducedpoints;
            if (gamestate.armor < 0)
                gamestate.armor = 0;
            DrawHealth ();
        }
    }

    if (points <= 0)
        return;

    gamestate.health -= points;

    if (gamestate.health<=0)
    {
        KillDamageType = damagetype;
        gamestate.health = 0;
        playstate = ex_died;
        killerobj = attacker;
    }

    if (useflash)
        StartDamageFlash (points);

    DrawHealth ();
    DrawFace ();
}

/*
===============
=
= HealSelf
=
===============
*/

boolean HealSelf (int points)
{
    if (gamestate.health >= 100)
        return false;
    gamestate.health += points;
    if (gamestate.health>100)
        gamestate.health = 100;

    DrawHealth ();
    DrawFace ();

    return true;
}

/*
===============
=
= GiveArmor
=
===============
*/

boolean GiveArmor (int points)
{
    if (gamestate.armor >= 100)
        return false;

    gamestate.armor += points;
    if (gamestate.armor>100)
        gamestate.armor = 100;
    DrawHealth ();

    return true;
}

//===========================================================================


/*
===============
=
= DrawLevel
=
===============
*/

void DrawLevel (void)
{
    if(viewsize == 21 && ingame) return;
#ifdef SPEAR
    if (gamestate.mapon == 20)
        LatchNumber (3,16,2,18);
    else
#endif
        LatchNumber (3,16,2,gamestate.mapon+1);
}

//===========================================================================


/*
===============
=
= DrawLives
=
===============
*/

void DrawLives (void)
{
    if(viewsize == 21 && ingame) return;
    LatchNumber (15,16,1,gamestate.lives);
}


/*
===============
=
= GiveExtraMan
=
===============
*/

void GiveExtraMan (void)
{
    if (gamestate.lives<9)
        gamestate.lives++;
    DrawLives ();
    PlaySoundGlobal(BONUS1UPSND,chan_item2,0);
}

//===========================================================================

/*
===============
=
= DrawScore
=
===============
*/

void DrawScore (void)
{
    if(viewsize == 21 && ingame) return;
    LatchNumber (6,16,7,gamestate.score);
}

/*
===============
=
= GivePoints
=
===============
*/

void GivePoints (int32_t points)
{
    gamestate.score += points;
    while (gamestate.score >= gamestate.nextextra)
    {
        gamestate.nextextra += EXTRAPOINTS;
        if (gamestate.lives < 9)
        {
            GetMessage("You earned an extra life!");
            GiveExtraMan ();
        }
    }
    DrawScore ();
}

/*
===============
=
= GiveSecret
=
===============
*/

void GiveSecret (void)
{
    if (gamestate.secretcount >= gamestate.secrettotal)
        return;
    gamestate.secretcount++;
    GetMessage("You found a secret!");
}

//===========================================================================

/*
==================
=
= DrawKeys
=
==================
*/

void DrawKeys (void)
{
    if(viewsize == 21 && ingame) return;

    StatusDrawPic (30,4,KEY0PIC);

    if (gamestate.keys & 1)
        StatusDrawPic (30,4,KEY1PIC);

    if (gamestate.keys & 2)
        StatusDrawPic (30,12,KEY2PIC);

    if (gamestate.keys & 4)
        StatusDrawPic (30,20,KEY3PIC);

    if (gamestate.keys & 8)
        StatusDrawPic (30,28,KEY4PIC);
}

/*
==================
=
= GiveWeapon
=
==================
*/

boolean GiveWeapon (int weapon, int ammo)
{
    ammo_t ammotype;

    if (weapon >= wp_mg42_mounted)
        return false;    // WSJ: mounted weapons can only be used, not picked up

    ammotype = CheckAmmoType(weapon);

    if (!gamestate.hasweapon[weapon])
    {   // WSJ: first time the player has picked up the weapon
        gamestate.hasweapon[weapon] = 1;
        if (gamestate.weapon < wp_mg42_mounted)
        {
            if ((weapontype)weapon == wp_sniper)
                gamestate.lastchosenweapon[weaponkey[weapon]] =
                gamestate.chosenweapon = wp_kar98k;
            else
                gamestate.lastchosenweapon[weaponkey[weapon]] =
                gamestate.chosenweapon = (weapontype)weapon;
        }
        if (!gamestate.attackframe)
            gamestate.weapon = gamestate.chosenweapon;

        if (gamestate.bestweapon<weapon)
            gamestate.bestweapon = (weapontype)weapon;

        GiveAmmo (ammo,ammotype);
        DrawAmmo ();
        return true;
    }

    return GiveAmmo (ammo,ammotype);
}

//===========================================================================

/*
===============
=
= DrawAmmo
=
===============
*/

void DrawAmmo (void)
{
    if(viewsize == 21 && ingame) return;
    if (CheckAmmoType(gamestate.weapon) == ammo_none)
        LatchNumber (26,16,3,-1);
    else
        LatchNumber (26,16,3,gamestate.ammo[CheckAmmoType(gamestate.weapon)]);
}

/*
===============
=
= GiveAmmo
=
===============
*/

boolean GiveAmmo (int ammo, ammo_t ammotype)
{
    if (!ammo || ammotype == ammo_none || gamestate.ammo[ammotype] >= maxammo[ammotype])
        return false;

    gamestate.ammo[ammotype] += ammo;
    if (gamestate.ammo[ammotype] > maxammo[ammotype])
        gamestate.ammo[ammotype] = maxammo[ammotype];
    if (gamestate.weapon!=gamestate.chosenweapon && WeaponAvailable(gamestate.chosenweapon)
        && !gamestate.attackframe && gamestate.weapon<wp_mg42_mounted)
        gamestate.weapon = gamestate.chosenweapon;

    DrawAmmo ();
    return true;
}

//===========================================================================

/*
==================
=
= GiveKey
=
==================
*/

void GiveKey (int key)
{
    gamestate.keys |= (1<<key);
    DrawKeys ();
    GetMessage(keyname[key]);
}

//===========================================================================

/*
==================
=
= StartInventory
=
= WSJ: set default health, ammo, etc.
= when player starts a new game or dies
=
==================
*/

void StartInventory (void)
{
    gamestate.health = 100;
    gamestate.armor = 0;
    gamestate.hasweapon[wp_knife] = gamestate.hasweapon[wp_pistol] = 1;
    gamestate.weapon = gamestate.bestweapon = gamestate.chosenweapon = wp_pistol;
    gamestate.ammo[ammo_bullet] = 8;
}

/*
=============================================================================

                                MOVEMENT

=============================================================================
*/

void DeleteBonus (statobj_t *check)
{
    check->flags = 0;
    check->itemnumber = 0;
    check->shapenum = -1;                   // remove from list
}

/*
===================
=
= GetBonus
=
===================
*/
void GetBonus (statobj_t *check)
{
    byte ammodiv = 1;
    int  points = 0;

    if (playstate == ex_died)
        return;    // WSJ: don't pick up anything while dying!

    if (check->flags & FL_USED)
        ammodiv = 2;    // WSJ: give half ammo if dropped by an enemy

    switch (check->itemnumber)
    {
        case    bo_firstaid:
            if (!HealSelf(25))
                return;
            PlayItemSound (HEALTH2SND);
            GetMessage ("First Aid Kit");
            break;

        case    bo_key1:
        case    bo_key2:
        case    bo_key3:
        case    bo_key4:
            GiveKey (check->itemnumber - bo_key1);
            PlayItemSound (GETKEYSND);
            break;

        case    bo_crown:
            points += 4000;
        case    bo_bible:
            points += 500;
        case    bo_chalice:
            points += 400;
        case    bo_cross:
            points += 100;
            PlayItemSound (BONUS1SND);
            sprintf (str, "Treasure: %u points", points);
            GetMessage (str);
            GivePoints (points);
            gamestate.treasurecount++;
            break;

        case    bo_clip:
            if (!GiveAmmo(8/ammodiv,ammo_bullet))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Pistol Clip");
            break;

        case    bo_clip_big:
            if (!GiveAmmo(32/ammodiv,ammo_bullet))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Submachine Gun Clip");
            break;

        case    bo_rounds:
            if (!GiveAmmo(5/ammodiv,ammo_rifle))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Rifle Rounds");
            break;

        case    bo_rounds_big:
            if (!GiveAmmo(30/ammodiv,ammo_rifle))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Box of Rifle Rounds");
            break;

        case    bo_stg44ammo:
            if (!GiveAmmo(30/ammodiv,ammo_stg44))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("STG44 Magazine");
            break;

        case    bo_fuel:
            if (!GiveAmmo(40/ammodiv,ammo_flame))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Fuel Tank");
            break;

        case    bo_ammobelt:
            if (!GiveAmmo(80/ammodiv,ammo_chaingun))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Ammo Belt");
            break;

        case    bo_rocket:
            if (!GiveAmmo(1,ammo_rocket))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Rocket");
            break;

        case    bo_flakammo:
            if (!GiveAmmo(20/ammodiv,ammo_flak))
                return;
            PlayItemSound (GETAMMOSND);
            GetMessage ("Box of Flak Shells");
            break;

        case    bo_pistol:
            if (!gamestate.hasweapon[wp_pistol])
                GiveWeapon (wp_pistol,8/ammodiv);
            else if (!GiveWeapon(wp_dualpistol,8/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Walther P38 Pistol");
            break;

        case    bo_mauser:
            if (!gamestate.hasweapon[wp_mauser])
                GiveWeapon (wp_mauser,20/ammodiv);
            else if (!GiveWeapon(wp_dualmauser,20/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Schnellfeuer Pistol");
            break;

        case    bo_kar98k:
            if (!GiveWeapon(wp_kar98k,5/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Kar98k Rifle");
            break;

        case    bo_kar98k_sniper:
            if (!GiveWeapon(wp_sniper,5/ammodiv))
                return;
            GiveWeapon (wp_kar98k,0);
            PlayItemSound (GETMACHINESND);
            GetMessage ("Kar98k Sniper Rifle");
            break;

        case    bo_mp40:
            if (!gamestate.hasweapon[wp_mp40])
                GiveWeapon (wp_mp40,32/ammodiv);
            else if (!GiveWeapon(wp_dualmp40,32/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("MP40 Submachine Gun");
            break;

        case    bo_stg44:
            if (!GiveWeapon(wp_stg44,30/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("STG44 Assault Rifle");
            break;

        case    bo_flamethrower:
            if (!GiveWeapon(wp_flamethrower,80/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Flamethrower");
            break;

        case    bo_mg42:
            if (!GiveWeapon(wp_mg42,80/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("MG42 Portable Machine Gun");
            break;

        case    bo_chaingun:
            if (!GiveWeapon(wp_chaingun,80/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Chaingun");
            break;

        case    bo_panzerschreck:
            if (!GiveWeapon(wp_panzerschreck,2/ammodiv))
                return;
            PlayItemSound (GETMACHINESND);
            GetMessage ("Panzerschreck");
            break;

        case    bo_flakgun_part:
            if (gamestate.flakgunparts >= NUMFLAKGUNPARTS)
                return;
            gamestate.flakgunparts++;
            PlayItemSound (GETMACHINESND);
            if (gamestate.flakgunparts >= NUMFLAKGUNPARTS)
            {
                GiveWeapon (wp_flakgun,100);
                GetMessage ("Flakgun completed!");
            }
            else
                GetMessage ("Flakgun Component");
            break;

        case    bo_armor_small:
            if (!GiveArmor(25))
                return;
            PlayItemSound (BONUS2SND);
            GetMessage ("Helmet");
            break;

        case    bo_armor_big:
            if (!GiveArmor(50))
                return;
            PlayItemSound (BONUS3SND);
            GetMessage ("Flak Jacket");
            break;

        case    bo_armor_uber:
            if (gamestate.armor >= 200)
                return;
            PlayItemSound (BONUS3SND);
            gamestate.armor = 200;
            DrawHealth ();
            GetMessage ("Uber-Armor");
            break;

        case    bo_armor_power:
            if (gamestate.gotregenarmor)
                return;
            PlaySoundGlobal(BONUS1UPSND,chan_item2,0);
            gamestate.gotregenarmor = true;
            GiveArmor (100);
            GetMessage ("Power Armor: your armor now regenerates!");
            break;

        case    bo_regensphere:
            if (gamestate.gotregenhealth)
                return;
            PlaySoundGlobal(BONUS1UPSND,chan_item2,0);
            gamestate.gotregenhealth = true;
            HealSelf (100);
            GetMessage ("Healing Orb: your health now regenerates!");
            break;

        case    bo_fullheal:
            HealSelf (100);
            GiveArmor (100);
            GiveExtraMan ();
            gamestate.treasurecount++;
            GetMessage ("1-UP!");
            break;

        case    bo_food:
            if (!HealSelf(10))
                return;
            PlayItemSound (HEALTH1SND);
            GetMessage ("Food");
            break;

        case    bo_alpo:
            if (!HealSelf(4))
                return;
            PlayItemSound (HEALTH1SND);
            GetMessage ("Dog Food");
            break;

        case    bo_acid:
            TakeDamage(1,NULL,dam_bullet,false,true);
        default:
            return;
    }

    StartBonusFlash ();
    DeleteBonus(check);
}

/*
===================
=
= TryMove
=
= returns true if move ok
= debug: use pointers to optimize
===================
*/

boolean TryMove (objtype *ob)
{
    int         xl,yl,xh,yh,x,y;
    objtype    *check;
    int32_t     deltax,deltay;

    xl = (ob->x-PLAYERSIZE) >>TILESHIFT;
    yl = (ob->y-PLAYERSIZE) >>TILESHIFT;

    xh = (ob->x+PLAYERSIZE) >>TILESHIFT;
    yh = (ob->y+PLAYERSIZE) >>TILESHIFT;

#define PUSHWALLMINDIST PLAYERSIZE

    //
    // check for solid walls
    //
    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (check && !ISPOINTER(check))
            {
                if(tilemap[x][y]==64 && x==pwallx && y==pwally)   // back of moving pushwall?
                {
                    switch(pwalldir)
                    {
                        case di_north:
                            if(ob->y-PUSHWALLMINDIST<=(pwally<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_west:
                            if(ob->x-PUSHWALLMINDIST<=(pwallx<<TILESHIFT)+((63-pwallpos)<<10))
                                return false;
                            break;
                        case di_east:
                            if(ob->x+PUSHWALLMINDIST>=(pwallx<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                        case di_south:
                            if(ob->y+PUSHWALLMINDIST>=(pwally<<TILESHIFT)+(pwallpos<<10))
                                return false;
                            break;
                    }
                }
                else return false;
            }
        }
    }

    //
    // check for actors
    //
    if (yl>0)
        yl--;
    if (yh<MAPSIZE-1)
        yh++;
    if (xl>0)
        xl--;
    if (xh<MAPSIZE-1)
        xh++;

    for (y=yl;y<=yh;y++)
    {
        for (x=xl;x<=xh;x++)
        {
            check = actorat[x][y];
            if (ISPOINTER(check) && check != player && (check->flags & FL_SHOOTABLE) )
            {
                deltax = ob->x - check->x;
                if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
                    continue;
                deltay = ob->y - check->y;
                if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
                    continue;

                return false;
            }
        }
    }

    return true;
}


/*
===================
=
= ClipMove
=
===================
*/

void ClipMove (objtype *ob, int32_t xmove, int32_t ymove)
{
    int32_t    basex,basey;

    basex = ob->x;
    basey = ob->y;

    ob->x = basex+xmove;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

#ifndef REMDEBUG
    if (noclip && ob->x > 2*TILEGLOBAL && ob->y > 2*TILEGLOBAL
        && ob->x < (((int32_t)(mapwidth-1))<<TILESHIFT)
        && ob->y < (((int32_t)(mapheight-1))<<TILESHIFT) )
        return;         // walk through walls
#endif

    ob->x = basex+xmove;
    ob->y = basey;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey+ymove;
    if (TryMove (ob))
        return;

    ob->x = basex;
    ob->y = basey;
}

//==========================================================================

/*
===================
=
= VictoryTile
=
===================
*/

void VictoryTile (void)
{
    playstate = ex_victorious;
}

/*
===================
=
= Thrust
=
===================
*/

// For player movement in demos exactly as in the original Wolf3D v1.4 source code
static fixed FixedByFracOrig(fixed a, fixed b)
{
    int sign = 0;
    if(b == 65536) b = 65535;
    else if(b == -65536) b = 65535, sign = 1;
    else if(b < 0) b = (-b), sign = 1;

    if(a < 0)
    {
        a = -a;
        sign = !sign;
    }
    fixed res = (fixed)(((int64_t) a * b) >> 16);
    if(sign)
        res = -res;
    return res;
}

void Thrust (int angle, int32_t speed)
{
    int32_t xmove,ymove;
    unsigned offset;


    //
    // ZERO FUNNY COUNTER IF MOVED!
    //
#ifdef SPEAR
    if (speed)
        funnyticount = 0;
#endif

    thrustspeed += speed;
    //
    // moving bounds speed
    //
    if (speed >= MINDIST*2)
        speed = MINDIST*2-1;

    xmove = DEMOCHOOSE_ORIG_SDL(
                FixedByFracOrig(speed, costable[angle]),
                FixedMul(speed,costable[angle]));
    ymove = DEMOCHOOSE_ORIG_SDL(
                -FixedByFracOrig(speed, sintable[angle]),
                -FixedMul(speed,sintable[angle]));

    ClipMove(player,xmove,ymove);

    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    offset = (player->tiley<<mapshift)+player->tilex;
    if ( (*(mapsegs[0] + offset)) >= AREATILE)
        player->areanumber = *(mapsegs[0] + offset) -AREATILE;
    if (!areabyplayer[player->areanumber] && player->areanumber<NUMAREAS)
        ConnectAreas ();    // WSJ: added so player can cross over to other area codes without using doors

    switch (*(mapsegs[1] + offset))
    {
        case EXITTILE:
            VictoryTile ();
            break;
        case ENDLEVELTILE:
            ExitLevel ();
            break;
    }
}


/*
=============================================================================

                                ACTIONS

=============================================================================
*/


/*
===============
=
= EnterMountedWeapon
=
= WSJ: player spawns at the position of the mounted
= weapon and takes control of it
=
===============
*/

void EnterMountedWeapon (weapontype weapon, int newx, int newy, int dir)
{
    int    facedir;

    switch (dir)
    {
    case di_east:
        facedir = EAST;
        break;
    case di_north:
        facedir = NORTH;
        break;
    case di_west:
        facedir = WEST;
        break;
    case di_south:
        facedir = SOUTH;
        break;
    }

    gamestate.chosenweapon = gamestate.weapon;
    if (gamestate.chosenweapon == wp_sniper)
        gamestate.chosenweapon = wp_kar98k;

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    SpawnPlayer(newx,newy,facedir);
    gamestate.weapon = weapon;
    DrawAmmo ();
    PlaySoundGlobal(USEMG42SND,chan_weapon_fire,0);
    GetMessage ("You now control the MG42!");
}

/*
===============
=
= ExitMountedWeapon
=
= WSJ: player leaves the mounted weapon and spawns at a
= position that's one tile away from it
=
===============
*/

#define CHECKEXITSPOT(x,y)                            \
{                                                     \
    blocked = false;                                  \
    check = actorat[x][y];                            \
    if (check && !ISPOINTER(check))                   \
        blocked = true;                               \
    if (ISPOINTER(check) && check != player           \
        && (check->flags & FL_SHOOTABLE))             \
        blocked = true;                               \
    if (!foundspot && !blocked)                       \
    {                                                 \
        newx = x;                                     \
        newy = y;                                     \
        foundspot = true;                             \
    }                                                 \
}

void ExitMountedWeapon (int dir)
{
    int        x,y,newx,newy,facedir;
    boolean    foundspot,blocked;
    objtype    *check;

    x = player->tilex;
    y = player->tiley;
    foundspot = false;

    switch (dir)
    {
    case di_east:
        facedir = EAST;
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x+1,y);
        break;
    case di_north:
        facedir = NORTH;
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y-1);
        break;
    case di_west:
        facedir = WEST;
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x,y+1);
        CHECKEXITSPOT(x-1,y);
        break;
    case di_south:
        facedir = SOUTH;
        CHECKEXITSPOT(x,y-1);
        CHECKEXITSPOT(x+1,y);
        CHECKEXITSPOT(x-1,y);
        CHECKEXITSPOT(x,y+1);
        break;
    }

    if (!foundspot)
    {
        GetMessage ("Not enough room to leave!");
        return;
    }

    buttonstate[bt_attack] = false;
    buttonheld[bt_attack] = false;
    gamestate.attackframe = 0;
    gamestate.weaponframe = 0;
    SpawnPlayer(newx,newy,facedir);
    gamestate.weapon = gamestate.chosenweapon;
    DrawAmmo ();
    actorat[x][y] = (objtype *) BLOCKTILE;          // consider it a blocking tile
    GetMessage ("You stopped using the weapon");
}

//===========================================================================


/*
===============
=
= Cmd_Fire
=
===============
*/

void Cmd_Fire (void)
{
    buttonheld[bt_attack] = true;

    gamestate.weaponframe = 0;

    player->state = &s_attack;

    gamestate.attackframe = 0;
    gamestate.attackcount =
        attackinfo[gamestate.weapon][gamestate.attackframe].tics;
    gamestate.weaponframe =
        attackinfo[gamestate.weapon][gamestate.attackframe].frame;
}

//===========================================================================

/*
===============
=
= Cmd_Use
=
===============
*/

void Cmd_Use (void)
{
    statobj_t *statptr;
    int       checkx,checky,dir,wallnum,side;

    //
    // find which cardinal direction the player is facing
    //
    if (player->angle < ANGLES/8 || player->angle > 7*ANGLES/8)
    {
        checkx = player->tilex + 1;
        checky = player->tiley;
        dir = di_east;
        side = WEST;
    }
    else if (player->angle < 3*ANGLES/8)
    {
        checkx = player->tilex;
        checky = player->tiley-1;
        dir = di_north;
        side = SOUTH;
    }
    else if (player->angle < 5*ANGLES/8)
    {
        checkx = player->tilex - 1;
        checky = player->tiley;
        dir = di_west;
        side = EAST;
    }
    else
    {
        checkx = player->tilex;
        checky = player->tiley + 1;
        dir = di_south;
        side = NORTH;
    }

    if (!buttonheld[bt_use] && gamestate.weapon >= wp_mg42_mounted
        && !gamestate.attackframe && !buttonstate[bt_attack])
    {
        //
        // WSJ: stop using MG42
        //
        buttonheld[bt_use] = true;
        ExitMountedWeapon (dir);
        return;
    }

    wallnum = tilemap[checkx][checky];

    if (wallnum && MAPSPOT(checkx,checky,1) == PUSHABLETILE)
    {
        //
        // pushable wall
        //
        for (statptr = &statobjlist[0]; statptr != laststatobj; statptr++)
        {
            if (statptr->tilex != checkx || statptr->tiley != checky)
                continue;

            if (statptr->itemnumber == bo_pushwall)
            {
                if (PushWall(checkx,checky,dir))
                {
                    GiveSecret();
                    DeleteBonus(statptr);
                }
                return;
            }
        }

        return;
    }

    if (buttonheld[bt_use])
        return;

    if (wallnum)
    {
        if (wallnum & 0x80)
        {
            buttonheld[bt_use] = true;
            OperateDoor (wallnum & ~0x80);
            return;
        }

        if (walltex[checkx][checky] == ELEVATORTILE)
        {
            buttonheld[bt_use] = true;
            walltex[checkx][checky]++;
            PlayItemSound(LEVELDONESND);
            ExitLevel();
            return;
        }
    }

    if (gamestate.attackframe || buttonstate[bt_attack])
        return;

    //
    // WSJ: use static object?
    //
    for (statptr = &statobjlist[0]; statptr != laststatobj; statptr++)
    {
        if (statptr->tilex != checkx || statptr->tiley != checky)
            continue;

        switch (statptr->itemnumber)
        {
            case bo_mg42_mounted:
                buttonheld[bt_use] = true;
                EnterMountedWeapon (wp_mg42_mounted,checkx,checky,dir);
                return;
        }
    }
}

/*
=============================================================================

                                PLAYER CONTROL

=============================================================================
*/



/*
===============
=
= SpawnPlayer
=
===============
*/

void SpawnPlayer (int tilex, int tiley, int dir)
{
    player->obclass = playerobj;
    player->active = ac_yes;
    player->tilex = tilex;
    player->tiley = tiley;
    player->areanumber = (byte) *(mapsegs[0]+(player->tiley<<mapshift)+player->tilex);
    player->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    player->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    player->state = &s_player;
    player->angle = (1-dir)*90;
    if (player->angle<0)
        player->angle += ANGLES;
    player->flags = FL_NEVERMARK;
    Thrust (0,0);                           // set some variables

    InitAreas ();
}


//===========================================================================

/*
===============
=
= T_KnifeAttack
=
= Update player hands, and try to do damage when the proper frame is reached
=
===============
*/

void    KnifeAttack (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;

    // actually fire
    dist = 0x7fffffff;
    closest = NULL;
    for (check=ob->next; check; check=check->next)
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && abs(check->viewx-centerx) < shootdelta)
        {
            if (check->transx < dist)
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (!closest || dist > 0x18000l)
    {
        // missed
        PlaySoundGlobal(ATKKNIFESND,chan_weapon_fire,0);
        return;
    }

    // hit something
    if (DamageActor (closest,player,1+(US_RndT()>>3),dam_melee))
        SpawnBloodPuff (closest, closest->x, closest->y);
    PlaySoundGlobal(HITENEMYSND,chan_weapon_fire,0);
}


/*
===============
=
= GunAttack
=
===============
*/

void    GunAttack (objtype *ob, int8_t attacknum)
{
    int weaponsound;

    switch (gamestate.weapon)
    {
        case wp_pistol:
            weaponsound = ATKPISTOLSND;
            FireBullet (ob,(10 + US_RndT()%25),dam_bullet,2,1,0);
            break;
        case wp_dualpistol:
            weaponsound = ATKPISTOLSND;
            FireBullet (ob,(10 + US_RndT()%25),dam_bullet,3,1,0);
            break;
        case wp_mauser:
            weaponsound = ATKMAUSERSND;
            FireBullet (ob,(6 + US_RndT()%20),dam_machinegun,4,1,0);
            break;
        case wp_dualmauser:
            weaponsound = ATKMAUSERSND;
            FireBullet (ob,(6 + US_RndT()%20),dam_machinegun,5,1,0);
            break;
        case wp_sniper:
            weaponsound = ATKKAR98SND;
            FireBullet (ob,(75 + US_RndT()%10),dam_bullet,0,1,0);
            break;
        case wp_kar98k:
            weaponsound = ATKKAR98SND;
            FireBullet (ob,(50 + US_RndT()%10),dam_bullet,0,1,0);
            break;
        case wp_mp40:
            weaponsound = ATKMP40SND;
            FireBullet (ob,(8 + US_RndT()%22),dam_bullet,2,1,0);
            break;
        case wp_dualmp40:
            weaponsound = ATKMP40SND;
            FireBullet (ob,(8 + US_RndT()%22),dam_bullet,4,1,0);
            break;
        case wp_stg44:
            weaponsound = ATKSTG44SND;
            FireBullet (ob,(25 + US_RndT()%20),dam_bullet,0,1,0);
            break;
        case wp_flamethrower:
            weaponsound = FLAMETHROWERSND;
            ProjFire (ob,bjfireobj,&s_fire1,0x3000l,0);
            break;
        case wp_mg42:
            weaponsound = ATKMG42SND;
            FireBullet (ob,(8 + US_RndT()%28),dam_machinegun,2,1,0);
            break;
        case wp_chaingun:
            weaponsound = ATKCHAINGUNSND;
            FireBullet (ob,(10 + US_RndT()%30),dam_chaingun,3,1,0);
            break;
        case wp_panzerschreck:
            weaponsound = MISSILEFIRESND;
            ProjFire (ob,bjrocketobj,&s_rocket,0x5000l,0);
            break;
        case wp_flakgun:
            weaponsound = ATKFLAKGUNSND;
            FireBullet (ob,100,dam_flak,3,1,(SPR_FLAKBOOM_1-SPR_BOOM_1));
            break;
        case wp_mg42_mounted:
            weaponsound = ATKMG42SND;
            FireBullet (ob,(12 + US_RndT()%28),dam_machinegun,2,1,0);
            break;
    }

    MakeNoise (player);

    if (attacknum != 5 && attacknum != 6)
        PlaySoundGlobal(weaponsound,chan_weapon_fire,0);
}


//
// WSJ: based on the original attack code
//
short AutoAim (objtype *ob)
{
    objtype *check,*closest;
    int32_t  dist;

    if (SNIPERMODE)
        return ob->angle;    // WSJ: don't auto-aim in sniper mode

    // WSJ: find potential targets
    dist = 0x7fffffff;
    closest = NULL;
    for (check=ob->next; check; check=check->next)
    {
        if ( (check->flags & FL_SHOOTABLE) && (check->flags & FL_VISABLE)
            && abs(check->viewx-centerx) < (shootdelta/2))
        {
            if (check->transx < dist && CheckLine(check))
            {
                dist = check->transx;
                closest = check;
            }
        }
    }

    if (closest != NULL)
    {
        // WSJ: calculate the angle required for the bullet to hit closest
        int32_t deltax,deltay;
        float   angle;
        short   iangle;

        deltax = closest->x - ob->x;
        deltay = ob->y - closest->y;
        angle = (float) atan2((float) deltay, (float) deltax);
        if (angle<0)
            angle = (float) (M_PI*2+angle);
        iangle = (short) (angle/(M_PI*2)*ANGLES);
        return iangle;
    }

    return ob->angle;    // WSJ: missed
}


//
// AlumiuN's new weaponry code - trace bullets
//
#define ENEMYCOLSIZE 0x5000l
#define PLAYERCOLSIZE 0x3000l    // WSJ: make the player harder to hit
#define STEPDIST 0x400l
#define MINTRACERDIST 0x36000l

void FireBullet (objtype *ob, short damage, damage_t damagetype, short spread, short numshots, short spriteoffset)
{
    int32_t         bulletx,bullety;
    short           btilex,btiley,iangle,bangle,dist;
    boolean         hit;
    objtype         *check;

    if (ob == player)
    {
        iangle = AutoAim (ob);
        dist = 0;
    }
    else
    {
        iangle = ob->angle;
        dist = OBJTILEDIST(ob,player);
    }

    for (int i = 0; i<numshots; i++)
    {
        bulletx = ob->x;
        bullety = ob->y;
        bangle = iangle;

        if (spread)
        {
            if (US_RndT() > 127)
                bangle -= US_RndT()%(spread + 1);
            else
                bangle += US_RndT()%(spread + 1);

            // WSJ: make sure it's a valid angle
            if (bangle<0)
                bangle+=ANGLES;
            else if (bangle>=ANGLES)
                bangle-=ANGLES;
        }

        hit = false;    // Assume the worst! Oh, and reset the variable. :)

        while(1)
        {
            bulletx += FixedMul(STEPDIST,costable[bangle]);
            bullety -= FixedMul(STEPDIST,sintable[bangle]);
            btilex = (short)(bulletx >> TILESHIFT);
            btiley = (short)(bullety >> TILESHIFT);

            // WSJ: added a tracer effect
            if (dist > 2)
            {
                int32_t deltax = player->x - bulletx;
                int32_t deltay = player->y - bullety;
                if (deltax > -MINTRACERDIST && deltax < MINTRACERDIST
                    && deltay > -MINTRACERDIST && deltay < MINTRACERDIST)
                    SpawnTracer(bulletx,bullety,bangle,spriteoffset);
            }

            //
            // check for solid walls
            //
            check = actorat[btilex][btiley];
            if (check && !ISPOINTER(check) && (uintptr_t)check != 64)
            {
                if ((uintptr_t)check < 128 || doorposition[(uintptr_t)check - 128] < 0xdfff/2)
                {
                    hit = true;
/*
                    switch (walltex[btilex][btiley])
                    {
                        case 75:
                            if (damagetype == dam_flak)
                                break;
                        case LANDSCAPETILE:
                            damagetype = dam_none;
                            break;
                    }
*/

                    switch (damagetype)
                    {
                        case dam_none:
                            break;
                        case dam_flak:
                            SpawnFlakPuff(ob,bulletx,bullety,damage,spriteoffset);
                            break;
                        default:
                            SpawnRicochet(bulletx,bullety,spriteoffset,damagetype);
                    }
                }
            }

            if (hit)
                break;

            if (ob == player)
            {
                for (check=player->next; check; check=check->next)
                {
                    if ( !(check->flags & FL_SHOOTABLE) )
                        continue;

                    if (LABS(bulletx-check->x) < ENEMYCOLSIZE && LABS(bullety-check->y) < ENEMYCOLSIZE)
                    {
                        hit = true;
                        switch (damagetype)
                        {
                            case dam_none:
                                break;
                            case dam_flak:
                                SpawnFlakPuff(ob,bulletx,bullety,damage,spriteoffset);
                                break;
                            default:
                                if (DamageActor (check, ob, damage, damagetype))
                                    SpawnBloodPuff (check, bulletx, bullety);
                                else
                                    SpawnRicochet(bulletx,bullety,spriteoffset,damagetype);
                        }
                    }

                    if (hit)
                        break;
                }
            }
            else if (LABS(bulletx-player->x) < PLAYERCOLSIZE && LABS(bullety-player->y) < PLAYERCOLSIZE)
            {
                hit = true;
                TakeDamage (damage, ob, damagetype, false, true);
            }

            if (hit)
                break;
        }
    }
}

//===========================================================================

/*
===============
=
= VictorySpin
=
===============
*/

void VictorySpin (void)
{
    int32_t    desty;

    if (player->angle > 270)
    {
        player->angle -= (short)(tics * 3);
        if (player->angle < 270)
            player->angle = 270;
    }
    else if (player->angle < 270)
    {
        player->angle += (short)(tics * 3);
        if (player->angle > 270)
            player->angle = 270;
    }

    desty = (((int32_t)player->tiley-5)<<TILESHIFT)-0x3000;

    if (player->y > desty)
    {
        player->y -= tics*4096;
        if (player->y < desty)
            player->y = desty;
    }
}


//===========================================================================

/*
===============
=
= T_Attack
=
===============
*/

void    T_Attack (objtype *ob)
{
    struct  atkinf  *cur;
    int     i;

    UpdateFace ();

    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    if ( buttonstate[bt_use] && !buttonheld[bt_use] )
        buttonstate[bt_use] = false;

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        buttonstate[bt_attack] = false;

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);

    //
    // change frame and fire
    //
    gamestate.attackcount -= (short) tics;
    while (gamestate.attackcount <= 0)
    {
        cur = &attackinfo[gamestate.weapon][gamestate.attackframe];
        switch (cur->attack)
        {
            case -3:
                if (WeaponAvailable(gamestate.weapon))
                    PlaySoundGlobal(KAR98BOLTSND,chan_weapon_bolt,0);
            case -2:
                if (WeaponAvailable(gamestate.weapon)) break;
            case -1:
                ob->state = &s_player;
                if (!HasAmmo(gamestate.weapon))
                {
                    if (gamestate.chosenweapon == wp_sniper)
                        gamestate.chosenweapon = wp_kar98k;
                    gamestate.weapon = BestAvailableWeapon(true);
                    if (gamestate.chosenweapon < gamestate.weapon)
                        gamestate.chosenweapon = gamestate.weapon;
                    DrawAmmo ();
                }
                else
                {
                    if (gamestate.weapon != gamestate.chosenweapon && gamestate.weapon < wp_mg42_mounted)
                    {
                        if (WeaponAvailable(gamestate.chosenweapon))
                            gamestate.weapon = gamestate.chosenweapon;
                        DrawAmmo ();
                    }
                }
                gamestate.attackframe = gamestate.weaponframe = 0;
                return;

            case 4:
            case 6:
                if (!HasAmmo(gamestate.weapon))
                    break;
                if (buttonstate[bt_attack])
                    gamestate.attackframe = gamestate.weaponframe = 0;
            case 1:
            case 5:
                if (!HasAmmo(gamestate.weapon))
                {       // can only happen with chain gun
                    gamestate.attackframe++;
                    break;
                }
                GunAttack (ob,cur->attack);
                if (!ammocheat && CheckAmmoType(gamestate.weapon) != ammo_none)
                    gamestate.ammo[CheckAmmoType(gamestate.weapon)]--;
                DrawAmmo ();
                break;

            case 2:
                KnifeAttack (ob);
                break;

            case 3:
                if (HasAmmo(gamestate.weapon) && buttonstate[bt_attack])
                    gamestate.attackframe = gamestate.weaponframe = 0;
                break;
        }

        gamestate.attackcount += cur->tics;
        gamestate.attackframe++;
        gamestate.weaponframe =
            attackinfo[gamestate.weapon][gamestate.attackframe].frame;
    }
}



//===========================================================================

/*
===============
=
= T_Player
=
===============
*/

void    T_Player (objtype *ob)
{
    if (gamestate.victoryflag)              // watching the BJ actor
    {
        VictorySpin ();
        return;
    }

    UpdateFace ();
    CheckWeaponChange ();

    if ( buttonstate[bt_use] )
        Cmd_Use ();

    if ( buttonstate[bt_attack] && !buttonheld[bt_attack])
        Cmd_Fire ();

    ControlMovement (ob);
    if (gamestate.victoryflag)              // watching the BJ actor
        return;

    plux = (word) (player->x >> UNSIGNEDSHIFT);                     // scale to fit in unsigned
    pluy = (word) (player->y >> UNSIGNEDSHIFT);
    player->tilex = (short)(player->x >> TILESHIFT);                // scale to tile values
    player->tiley = (short)(player->y >> TILESHIFT);
}
