// WL_GAME.C

#include <math.h>
#include "wl_def.h"
#include <SDL_mixer.h>
#pragma hdrstop

#ifdef MYPROFILE
#include <TIME.H>
#endif


/*
=============================================================================

                             LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

                             GLOBAL VARIABLES

=============================================================================
*/

boolean         ingame,fizzlein;
gametype        gamestate;
byte            bordercol=VIEWCOLOR;        // color of the Change View/Ingame border

byte            searchdata[0x10000];
byte*           sdqstart = searchdata + maparea;

#ifdef SPEAR
int32_t         spearx,speary;
unsigned        spearangle;
boolean         spearflag;
#endif

#ifdef USE_FEATUREFLAGS
int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight;
#endif

//
// ELEVATOR BACK MAPS - REMEMBER (-1)!!
//
// WSJ: no longer used - now it's determined by player's start areanumber
//
int ElevatorBackTo[]={1,1,7,3,5,3};

void SetupGameLevel (void);
void DrawPlayScreen (void);
void LoadLatchMem (void);
void GameLoop (void);

#ifdef DEBUGKEYS
boolean infinitelives;    // WSJ: used for tedlevel
#endif

/*
=============================================================================

                             LOCAL VARIABLES

=============================================================================
*/


/*
==========================
=
= PlaceValidArea
=
==========================
*/

void PlaceValidArea (int tilex, int tiley)
{
    word *map,tile;
    map = mapsegs[0]+(tiley<<mapshift)+tilex;
    tile = *map;

    if (VALIDAREATILE(tile))
        return;

    if (VALIDAREATILE(*(map+1)))
        tile = *(map+1);
    else if (VALIDAREATILE(*(map-MAPSIZE)))
        tile = *(map-MAPSIZE);
    else if (VALIDAREATILE(*(map+MAPSIZE)))
        tile = *(map+MAPSIZE);
    else if (VALIDAREATILE(*(map-1)))
        tile = *(map-1);
    else    // WSJ: if all else fails, use the first floor code
        tile = AREATILE;

    *map = tile;
}


/*
==========================
=
= ScanInfoPlane
=
= Spawn all actors and mark down special places
=
==========================
*/

// WSJ: quick fix for the legacy enemy directions
static const byte legacydir[4] = {1,0,3,2};

static void ScanInfoPlane(void)
{
    unsigned x,y;
    int      tile;
    word     *start;
    boolean  flakavailable = gamestate.hasweapon[wp_flakgun];

    start = mapsegs[1];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *start++;
            if (!tile)
                continue;

            switch (tile)
            {
                case 19:
                case 20:
                case 21:
                case 22:
                    SpawnPlayer(x,y,NORTH+tile-19);
                    if (loadedgame)
                        break;
                    if (gamestate.mapon > LRpack)          // WSJ: secret level?
                    {
                        if (player->areanumber <= LRpack)
                            gamestate.nextmap = player->areanumber;
                        else
                            gamestate.nextmap = 0;
                    }
                    else if (gamestate.mapon == LRpack)    // WSJ: last level?
                        gamestate.nextmap = 0;
                    else
                        gamestate.nextmap = gamestate.mapon+1;
                    break;

                case 23:
                case 24:
                case 25:
                case 26:
                case 27:
                case 28:
                case 29:
                case 30:

                case 31:
                case 32:
                case 33:
                case 34:
                case 35:
                case 36:
                case 37:
                case 38:

                case 39:
                case 40:
                case 41:
                case 42:
                case 43:
                case 44:
                case 45:
                case 46:

                case 47:
                case 48:
                case 49:
                case 50:
                case 51:
                case 52:
                case 53:
                case 54:

                case 55:
                case 56:
                case 57:
                case 58:
                case 59:
                case 60:
                case 61:
                case 62:

                case 63:
                case 64:
                case 65:
                case 66:
                case 67:
                case 68:
                case 69:
                case 70:
                case 71:
                case 72:    // Scoped Kar98k
                case 73:    // STG44
                case 74:    // rifle rounds
                case 75:    // box of rifle rounds
                case 76:    // large clip
                case 77:    // operating table 1
                case 78:    // operating table 2
                case 79:    // operating table 3
                case 80:    // bronze key
                case 81:    // elevator key
                case 82:    // toxic barrel
                case 83:    // dead prisoner
                case 84:    // brains
                case 85:    // severed arm
                case 86:    // gibs 2
                case 87:    // gibs 3
                case 88:    // Pistol
                case 89:    // lab table 2

                    SpawnStatic(x,y,tile-23);
                    break;

                case 1:     // Flamethrower
                case 2:     // fuel canister
                case 3:     // Chaingun
                case 4:     // ammo belt
                case 5:     // Panzerschreck
                case 6:     // rocket
                case 7:     // helmet
                case 8:     // flak jacket
                case 9:     // Uber-Armor
                case 10:    // Kar98k
                case 11:    // mounted MG42
                case 12:    // sandbags 1 (vertical)
                case 13:    // sandbags 1 (horizontal)
                case 14:    // sandbags 2 (vertical)
                case 15:    // sandbags 2 (horizontal)
                case 16:    // concrete window (vertical)
                case 17:    // concrete window (horizontal)
                case 18:    // charred skeleton

                    SpawnStatic(x,y,tile+66);
                    break;

                case 352:
                    SpawnSpecialStatic(x,y,block,SPR_STAT2_33_1,FL_DIR_VERT_MID,0);
                    break;
                case 353:
                    SpawnSpecialStatic(x,y,block,SPR_STAT2_33_2,FL_DIR_HORIZ_MID,0);
                    break;
                case 360:
                    if (!gamestate.flakgunparts)
                        SpawnSpecialStatic(x,y,bo_flakgun_part,SPR_STAT2_35,0,0);
                    break;
                case 361:
                    if (gamestate.flakgunparts == 1)
                        SpawnSpecialStatic(x,y,bo_flakgun_part,SPR_STAT2_36,0,0);
                    break;
                case 362:
                    if (gamestate.flakgunparts == 2)
                    {
                        SpawnSpecialStatic(x,y,bo_flakgun_part,SPR_STAT2_36,0,0);
                        flakavailable = true;
                    }
                    break;
                case 364:
                    if (!gamestate.gotregenarmor)
                        SpawnSpecialStatic(x,y,bo_armor_power,SPR_STAT2_38,FL_FULLBRIGHT,0);
                    break;
                case 365:
                    if (!gamestate.gotregenhealth)
                        SpawnSpecialStatic(x,y,bo_regensphere,SPR_STAT2_39,FL_FULLBRIGHT,0);
                    break;

                case 366:   // truck back
                case 367:   // truck middle
                case 368:   // truck front
                case 369:   // STG44 ammo

                    SpawnStatic(x,y,tile-281);
                    break;

//
// P wall
//
                case 98:
                    if (!loadedgame)
                        gamestate.secrettotal++;
                    PlaceValidArea(x,y);
                    SpawnSpecialStatic(x,y,bo_pushwall,0,0,0);
                    break;

//
// guard
//
                case 180:
                case 181:
                case 182:
                case 183:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 144:
                case 145:
                case 146:
                case 147:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 108:
                case 109:
                case 110:
                case 111:
                    tile -= 108;
                    switch (Random(6))
                    {
                        case 0:
                            SpawnStand(guardobj,x,y,tile,0);
                            break;
                        case 1:
                            SpawnStand(guardobj,x,y,tile,(SPR_GRD2_S_1-SPR_GRD_S_1));
                            break;
                        case 2:
                        case 3:
                            SpawnStand(riflemanobj,x,y,tile,(SPR_RIFLE_S_1-SPR_GRD_S_1));
                            break;
                        case 4:
                            SpawnStand(infantryobj,x,y,tile,(SPR_INF_S_1-SPR_GRD_S_1));
                            break;
                        case 5:
                            SpawnStand(infantryobj,x,y,tile,(SPR_INF2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;


                case 184:
                case 185:
                case 186:
                case 187:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 148:
                case 149:
                case 150:
                case 151:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 112:
                case 113:
                case 114:
                case 115:
                    tile -= 112;
                    switch (Random(6))
                    {
                        case 0:
                            SpawnPatrol(guardobj,x,y,tile,0);
                            break;
                        case 1:
                            SpawnPatrol(guardobj,x,y,tile,(SPR_GRD2_S_1-SPR_GRD_S_1));
                            break;
                        case 2:
                        case 3:
                            SpawnPatrol(riflemanobj,x,y,tile,(SPR_RIFLE_S_1-SPR_GRD_S_1));
                            break;
                        case 4:
                            SpawnPatrol(infantryobj,x,y,tile,(SPR_INF_S_1-SPR_GRD_S_1));
                            break;
                        case 5:
                            SpawnPatrol(infantryobj,x,y,tile,(SPR_INF2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;

//
// dead enemies
//
                case 124:
                    SpawnSimpleStatic (x,y,SPR_GRD_DEAD);
                    break;
                case 125:
                    SpawnSimpleStatic (x,y,SPR_SNIPER_DEAD);
                    break;
                case 161:
                    SpawnSimpleStatic (x,y,SPR_STRANGLEDGUARD);
                    break;
                case 344:
                    SpawnSimpleStatic (x,y,SPR_DOG_DEAD);
                    break;
                case 345:
                    SpawnSimpleStatic (x,y,SPR_SS_DEAD);
                    break;
                case 346:
                    SpawnSimpleStatic (x,y,SPR_OFC_DEAD);
                    break;
                case 347:
                    SpawnSimpleStatic (x,y,SPR_MUT_DEAD);
                    break;
                case 348:
                    SpawnSimpleStatic (x,y,SPR_SCI_DEAD);
                    break;
                case 349:
                    SpawnSimpleStatic (x,y,SPR_SSOFC_DEAD);
                    break;
                case 350:
                    SpawnSimpleStatic (x,y,SPR_WAFFEN_DEAD);
                    break;
                case 351:
                    SpawnSimpleStatic (x,y,SPR_WAFFENOFC_DEAD);
                    break;

//
// officer
//
                case 188:
                case 189:
                case 190:
                case 191:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 152:
                case 153:
                case 154:
                case 155:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 116:
                case 117:
                case 118:
                case 119:
                    tile -= 116;
                    switch (Random(3))
                    {
                        case 0:
                        case 1:
                            SpawnStand(officerobj,x,y,tile,(SPR_OFC_S_1-SPR_GRD_S_1));
                            break;
                        case 2:
                            SpawnStand(officer2obj,x,y,tile,(SPR_OFC2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;


                case 192:
                case 193:
                case 194:
                case 195:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 156:
                case 157:
                case 158:
                case 159:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 120:
                case 121:
                case 122:
                case 123:
                    tile -= 120;
                    switch (Random(3))
                    {
                        case 0:
                        case 1:
                            SpawnPatrol(officerobj,x,y,tile,(SPR_OFC_S_1-SPR_GRD_S_1));
                            break;
                        case 2:
                            SpawnPatrol(officer2obj,x,y,tile,(SPR_OFC2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;


//
// ss
//
                case 198:
                case 199:
                case 200:
                case 201:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 162:
                case 163:
                case 164:
                case 165:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 126:
                case 127:
                case 128:
                case 129:
                    tile -= 126;
                    switch (Random(2))
                    {
                        case 0:
                            SpawnStand(ssobj,x,y,tile,(SPR_SS_S_1-SPR_GRD_S_1));
                            break;
                        case 1:
                            SpawnStand(ssriflemanobj,x,y,tile,(SPR_SSRIFLE_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;


                case 202:
                case 203:
                case 204:
                case 205:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 166:
                case 167:
                case 168:
                case 169:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 130:
                case 131:
                case 132:
                case 133:
                    tile -= 130;
                    switch (Random(2))
                    {
                        case 0:
                            SpawnPatrol(ssobj,x,y,tile,(SPR_SS_S_1-SPR_GRD_S_1));
                            break;
                        case 1:
                            SpawnPatrol(ssriflemanobj,x,y,tile,(SPR_SSRIFLE_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;

//
// dogs
//
                case 206:
                case 207:
                case 208:
                case 209:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 170:
                case 171:
                case 172:
                case 173:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 134:
                case 135:
                case 136:
                case 137:
                    tile -= 134;
                    switch (Random(2))
                    {
                        case 0:
                            SpawnStand(dogobj,x,y,tile,0);
                            break;
                        case 1:
                            SpawnStand(dogobj,x,y,tile,(SPR_DOG2_W1_1-SPR_DOG_W1_1));
                            break;
                    }
                    break;


                case 210:
                case 211:
                case 212:
                case 213:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 36;
                case 174:
                case 175:
                case 176:
                case 177:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 36;
                case 138:
                case 139:
                case 140:
                case 141:
                    tile -= 138;
                    switch (Random(2))
                    {
                        case 0:
                            SpawnPatrol(dogobj,x,y,tile,0);
                            break;
                        case 1:
                            SpawnPatrol(dogobj,x,y,tile,(SPR_DOG2_W1_1-SPR_DOG_W1_1));
                            break;
                    }
                    break;

//
// boss
//
                case 160:
                    SpawnStand (flametrooperobj,x,y,nodir,(SPR_FAKE_W1-SPR_BOSS_W1));
                    break;
                case 178:
                    SpawnStand (deatharmorobj,x,y,nodir,(SPR_MECHA_W1-SPR_BOSS_W1));
                    break;
                case 179:
                    SpawnStand (fatobj,x,y,nodir,(SPR_FAT_W1-SPR_BOSS_W1));
                    break;
                case 196:
                    SpawnStand (schabbobj,x,y,nodir,(SPR_SCHABB_W1-SPR_BOSS_W1));
                    break;
                case 197:
                    SpawnStand (chaingunnerobj,x,y,nodir,(SPR_GRETEL_W1-SPR_BOSS_W1));
                    break;
                case 214:
                    SpawnStand (bossobj,x,y,nodir,0);
                    break;
                case 215:
                    SpawnStand (rockettrooperobj,x,y,nodir,(SPR_GIFT_W1-SPR_BOSS_W1));
                    break;
                case 250:
                    SpawnStand (panzobj,x,y,nodir,(SPR_PANZ_W1-SPR_BOSS_W1));
                    break;

//
// mutants
//
                case 252:
                case 253:
                case 254:
                case 255:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 18;
                case 234:
                case 235:
                case 236:
                case 237:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 18;
                case 216:
                case 217:
                case 218:
                case 219:
                    switch (Random(2))
                    {
                        case 0:
                            SpawnStand(mutantobj,x,y,tile-216,(SPR_MUT_S_1-SPR_GRD_S_1));
                            break;
                        case 1:
                            SpawnStand(mutantobj,x,y,tile-216,(SPR_MUT2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;

                case 256:
                case 257:
                case 258:
                case 259:
                    if (gamestate.difficulty<gd_hard)
                        break;
                    tile -= 18;
                case 238:
                case 239:
                case 240:
                case 241:
                    if (gamestate.difficulty<gd_medium)
                        break;
                    tile -= 18;
                case 220:
                case 221:
                case 222:
                case 223:
                    switch (Random(2))
                    {
                        case 0:
                            SpawnPatrol(mutantobj,x,y,tile-220,(SPR_MUT_S_1-SPR_GRD_S_1));
                            break;
                        case 1:
                            SpawnPatrol(mutantobj,x,y,tile-220,(SPR_MUT2_S_1-SPR_GRD_S_1));
                            break;
                    }
                    break;

//
// misc. actors
//
                case 224:
                    SpawnStand (sniperobj,x,y,nodir,(SPR_SNIPER_PAIN_1-SPR_GRD_PAIN_1));
                    break;
                case 225:
                    SpawnSpecialStatic (x,y,bo_acid,SPR_STAT_53,0,0);
                    break;
                case 226:
                    SpawnSpecialStatic (x,y,block,SPR_STAT_52_1,0,2);
                    break;
                case 227:
                    SpawnElectrode(x,y);
                    break;

//
// WSJ: gunners
//
                case 228:
                    SpawnStand (mg42guardobj,x,y,nodir,(SPR_GRD_MG42_PAIN_1-SPR_GRD_PAIN_1));
                    break;
                case 229:
                    SpawnStand (mg42ssobj,x,y,nodir,(SPR_SS_MG42_PAIN_1-SPR_GRD_PAIN_1));
                    break;
                case 230:
                    SpawnStand (mg42waffenssobj,x,y,nodir,(SPR_WAFFEN_MG42_PAIN_1-SPR_GRD_PAIN_1));
                    break;

//
// WSJ: legacy enemy spawns
//
                case 324:
                case 325:
                case 326:
                case 327:
                    if (US_RndT() < 128)
                        SpawnStand (ssofficerobj,x,y,legacydir[tile-324],(SPR_SSOFC_S_1-SPR_GRD_S_1));
                    else
                        SpawnStand (ssofficer2obj,x,y,legacydir[tile-324],(SPR_SSOFC2_S_1-SPR_GRD_S_1));
                    break;
                case 328:
                    if (US_RndT() < 128)
                        SpawnPatrol (ssofficerobj,x,y,0,(SPR_SSOFC_S_1-SPR_GRD_S_1));
                    else
                        SpawnPatrol (ssofficer2obj,x,y,0,(SPR_SSOFC2_S_1-SPR_GRD_S_1));
                    break;

                case 329:
                case 330:
                case 331:
                case 332:
                    if (US_RndT() < 128)
                        SpawnStand (waffenssobj,x,y,legacydir[tile-329],(SPR_WAFFEN_S_1-SPR_GRD_S_1));
                    else
                        SpawnStand (waffenssobj,x,y,legacydir[tile-329],(SPR_WAFFEN2_S_1-SPR_GRD_S_1));
                    break;
                case 333:
                    if (US_RndT() < 128)
                        SpawnPatrol (waffenssobj,x,y,0,(SPR_WAFFEN_S_1-SPR_GRD_S_1));
                    else
                        SpawnPatrol (waffenssobj,x,y,0,(SPR_WAFFEN2_S_1-SPR_GRD_S_1));
                    break;

                case 334:
                case 335:
                case 336:
                case 337:
                    SpawnStand (waffenofficerobj,x,y,legacydir[tile-334],(SPR_WAFFENOFC_S_1-SPR_GRD_S_1));
                    break;
                case 338:
                    SpawnPatrol (waffenofficerobj,x,y,0,(SPR_WAFFENOFC_S_1-SPR_GRD_S_1));
                    break;

                case 339:
                case 340:
                case 341:
                case 342:
                    if (US_RndT() < 128)
                        SpawnStand (scientistobj,x,y,legacydir[tile-339],(SPR_SCI_S_1-SPR_GRD_S_1));
                    else
                        SpawnStand (scientistobj,x,y,legacydir[tile-339],(SPR_SCI2_S_1-SPR_GRD_S_1));
                    break;
                case 343:
                    if (US_RndT() < 128)
                        SpawnPatrol (scientistobj,x,y,0,(SPR_SCI_S_1-SPR_GRD_S_1));
                    else
                        SpawnPatrol (scientistobj,x,y,0,(SPR_SCI2_S_1-SPR_GRD_S_1));
                    break;

                case 354:
                case 355:
                case 356:
                case 357:
                    SpawnStand (waffenssobj,x,y,legacydir[tile-354],(SPR_WAFFENGAS_S_1-SPR_GRD_S_1));
                    break;
                case 358:
                    SpawnPatrol (waffenssobj,x,y,0,(SPR_WAFFENGAS_S_1-SPR_GRD_S_1));
                    break;

                case 359:
                    SpawnStand (mechtrooperobj,x,y,nodir,0);
                    break;

//
// WSJ: ambient sound objects
//
                case 600:
                    if (!loadedgame && VALIDAREATILE(MAPSPOT(x,y,0)))
                        gamestate.ambience[AreaAt(x,y)] = amb_none;
                    break;
            }

            if (!loadedgame && tile > 600 && tile < 700 && VALIDAREATILE(MAPSPOT(x,y,0)))
                gamestate.ambience[AreaAt(x,y)] = tile-600;
            else if (tile > 799 && tile < 900)
                PlaceValidArea(x,y);
        }
    }

//
// WSJ: if Flakgun or its last part is present, spawn ammo for it
//
    if (!flakavailable)
        return;

    start = mapsegs[1];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *start++;

            if (tile == 363)
                SpawnSpecialStatic(x,y,bo_flakammo,SPR_STAT2_37,0,0);
        }
    }
}

//==========================================================================

/*
==================
=
= SetupGameLevel
=
==================
*/

void SetupGameLevel (void)
{
    int  x,y;
    word *map;
    word tile;


    if (!loadedgame)
    {
        gamestate.TimeCount
            = gamestate.killcount = gamestate.killtotal
            = gamestate.secretcount = gamestate.secrettotal
            = gamestate.treasurecount = gamestate.treasuretotal
            = gamestate.attackframe = gamestate.weaponframe
            = pwallstate = pwallpos = facetimes = 0;
        gamestate.ambience[0] = amb_none;
        for (int i=1; i<NUMAREAS; i++) { gamestate.ambience[i]=amb_default; }
        LastAttacker = NULL;
        killerobj = NULL;
    }

    alertedarea = lastalertedarea = 0;

    GetMessage("");
    GetSubtitle("");

    if (demoplayback || demorecord)
        US_InitRndT (false);
    else
        US_InitRndT (true);

//
// load the level
//
    CA_CacheMap (gamestate.mapon+10*gamestate.episode);
    mapon-=gamestate.episode*10;

#ifdef USE_FEATUREFLAGS
    // Temporary definition to make things clearer
    #define MXX MAPSIZE - 1

    // Read feature flags data from map corners and overwrite corners with adjacent tiles
    ffDataTopLeft     = MAPSPOT(0,   0,   0); MAPSPOT(0,   0,   0) = MAPSPOT(1,       0,       0);
    ffDataTopRight    = MAPSPOT(MXX, 0,   0); MAPSPOT(MXX, 0,   0) = MAPSPOT(MXX,     1,       0);
    ffDataBottomRight = MAPSPOT(MXX, MXX, 0); MAPSPOT(MXX, MXX, 0) = MAPSPOT(MXX - 1, MXX,     0);
    ffDataBottomLeft  = MAPSPOT(0,   MXX, 0); MAPSPOT(0,   MXX, 0) = MAPSPOT(0,       MXX - 1, 0);

    #undef MXX
#endif

//
// copy the wall data to a data segment array
//
    memset (tilemap,0,sizeof(tilemap));
    memset (walltex,0,sizeof(walltex));
    memset (actorat,0,sizeof(actorat));
    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *map++;
            if (!tile)
                continue;

            if (tile<90)
            {
                // solid wall
                tilemap[x][y] = 1;
                actorat[x][y] = (objtype *)(uintptr_t) 1;
                walltex[x][y] = tile;
            }
        }
    }

//
// spawn doors
//
    InitActorList ();                       // start spawning things with a clean slate
    InitDoorList ();
    InitStaticList ();

    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *map++;

            if (tile > 89 && tile < 106)
                SpawnDoor (x,y,!(tile&1),(tile-90)/2);
            else if (tile > 255 && tile < 268)
                SpawnDoor (x,y,!(tile&1),(tile-240)/2);
        }
    }

//
// spawn actors
//
    ScanInfoPlane ();


//
// take out the ambush markers
//
    map = mapsegs[0];
    for (y=0;y<mapheight;y++)
    {
        for (x=0;x<mapwidth;x++)
        {
            tile = *map++;

            if (tile && !tilemap[x][y])
                PlaceValidArea(x,y);
        }
    }
}


//==========================================================================


/*
===================
=
= DrawPlayBorderSides
=
= To fix window overwrites
=
===================
*/
void DrawPlayBorderSides(void)
{
    if(viewsize == 21) return;

	const int sw = screenWidth;
	const int sh = screenHeight;
	const int vw = viewwidth;
	const int vh = viewheight;
	const int px = scaleFactor; // size of one "pixel"

	const int h  = sh - px * STATUSLINES;
	const int xl = sw / 2 - vw / 2;
	const int yl = (h - vh) / 2;

    if(xl != 0)
    {
	    VWB_BarScaledCoord(0,            0, xl - px,     h, bordercol);                 // left side
	    VWB_BarScaledCoord(xl + vw + px, 0, xl - px * 2, h, bordercol);                 // right side
    }

    if(yl != 0)
    {
	    VWB_BarScaledCoord(0, 0,            sw, yl - px, bordercol);                    // upper side
	    VWB_BarScaledCoord(0, yl + vh + px, sw, yl - px, bordercol);                    // lower side
    }

    if(xl != 0)
    {
        // Paint game view border lines
	    VWB_BarScaledCoord(xl - px, yl - px, vw + px, px,          0);                      // upper border
	    VWB_BarScaledCoord(xl,      yl + vh, vw + px, px,          bordercol - 2);          // lower border
	    VWB_BarScaledCoord(xl - px, yl - px, px,      vh + px,     0);                      // left border
	    VWB_BarScaledCoord(xl + vw, yl - px, px,      vh + px * 2, bordercol - 2);          // right border
	    VWB_BarScaledCoord(xl - px, yl + vh, px,      px,          bordercol - 3);          // lower left highlight
    }
    else
    {
        // Just paint a lower border line
        VWB_BarScaledCoord(0, yl+vh, vw, px, bordercol-2);       // lower border
    }
}


/*
===================
=
= DrawStatusBorder
=
===================
*/

void DrawStatusBorder (byte color)
{
    int statusborderw = (screenWidth-scaleFactor*320)/2;

    VWB_BarScaledCoord (0,0,screenWidth,screenHeight-scaleFactor*(STATUSLINES-3),color);
    VWB_BarScaledCoord (0,screenHeight-scaleFactor*(STATUSLINES-3),
        statusborderw+scaleFactor*8,scaleFactor*(STATUSLINES-4),color);
    VWB_BarScaledCoord (0,screenHeight-scaleFactor*2,screenWidth,scaleFactor*2,color);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*8, screenHeight-scaleFactor*(STATUSLINES-3),
        statusborderw+scaleFactor*8,scaleFactor*(STATUSLINES-4),color);

    VWB_BarScaledCoord (statusborderw+scaleFactor*9, screenHeight-scaleFactor*3,
        scaleFactor*97, scaleFactor*1, color-1);
    VWB_BarScaledCoord (statusborderw+scaleFactor*106, screenHeight-scaleFactor*3,
        scaleFactor*161, scaleFactor*1, color-2);
    VWB_BarScaledCoord (statusborderw+scaleFactor*267, screenHeight-scaleFactor*3,
        scaleFactor*44, scaleFactor*1, color-3);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*9, screenHeight-scaleFactor*(STATUSLINES-4),
        scaleFactor*1, scaleFactor*20, color-2);
    VWB_BarScaledCoord (screenWidth-statusborderw-scaleFactor*9, screenHeight-scaleFactor*(STATUSLINES/2-4),
        scaleFactor*1, scaleFactor*14, color-3);
}


/*
===================
=
= DrawPlayBorder
=
===================
*/

void DrawPlayBorder (void)
{
	const int px = scaleFactor; // size of one "pixel"

    if (bordercol != VIEWCOLOR)
        DrawStatusBorder(bordercol);
    else
    {
        const int statusborderw = (screenWidth-px*320)/2;
        VWB_BarScaledCoord (0, screenHeight-px*STATUSLINES,
            statusborderw+px*8, px*STATUSLINES, bordercol);
        VWB_BarScaledCoord (screenWidth-statusborderw-px*8, screenHeight-px*STATUSLINES,
            statusborderw+px*8, px*STATUSLINES, bordercol);
    }

    if((unsigned) viewheight == screenHeight) return;

    VWB_BarScaledCoord (0,0,screenWidth,screenHeight-px*STATUSLINES,bordercol);

    const int xl = screenWidth/2-viewwidth/2;
    const int yl = (screenHeight-px*STATUSLINES-viewheight)/2;
    VWB_BarScaledCoord (xl,yl,viewwidth,viewheight,0);

    if(xl != 0)
    {
        // Paint game view border lines
        VWB_BarScaledCoord(xl-px, yl-px, viewwidth+px, px, 0);                      // upper border
        VWB_BarScaledCoord(xl, yl+viewheight, viewwidth+px, px, bordercol-2);       // lower border
        VWB_BarScaledCoord(xl-px, yl-px, px, viewheight+px, 0);                     // left border
        VWB_BarScaledCoord(xl+viewwidth, yl-px, px, viewheight+2*px, bordercol-2);  // right border
        VWB_BarScaledCoord(xl-px, yl+viewheight, px, px, bordercol-3);              // lower left highlight
    }
    else
    {
        // Just paint a lower border line
        VWB_BarScaledCoord(0, yl+viewheight, viewwidth, px, bordercol-2);       // lower border
    }
}


/*
===================
=
= DrawPlayScreen
=
===================
*/

void DrawPlayScreen (void)
{
    VWB_DrawPicScaledCoord ((screenWidth-scaleFactor*320)/2,screenHeight-scaleFactor*STATUSLINES,STATUSBARPIC);
    DrawPlayBorder ();

    DrawFace ();
    DrawHealth ();
    DrawLives ();
    DrawLevel ();
    DrawAmmo ();
    DrawKeys ();
    DrawScore ();
}

// Uses LatchDrawPic instead of StatusDrawPic
void LatchNumberHERE (int x, int y, unsigned width, int32_t number)
{
    unsigned length,c;
    char str[20];

    ltoa (number,str,10);

    length = (unsigned) strlen (str);

    while (length<width)
    {
        LatchDrawPic (x,y,N_BLANKPIC);
        x++;
        width--;
    }

    c = length <= width ? 0 : length-width;

    while (c<length)
    {
        LatchDrawPic (x,y,str[c]-'0'+ N_0PIC);
        x++;
        c++;
    }
}

void ShowActStatus()
{
    // Draw status bar without borders
    byte *source = grsegs[STATUSBARPIC];
    int	picnum = STATUSBARPIC - STARTPICS;
    int width = pictable[picnum].width;
    int height = pictable[picnum].height;
    int destx = (screenWidth-scaleFactor*320)/2 + 9 * scaleFactor;
    int desty = screenHeight - (height - 4) * scaleFactor;
    VL_MemToScreenScaledCoord(source, width, height, 9, 4, destx, desty, width - 18, height - 7);

    ingame = false;
    DrawFace ();
    DrawHealth ();
    DrawLives ();
    DrawLevel ();
    DrawAmmo ();
    DrawKeys ();
    DrawScore ();
    ingame = true;
}


//==========================================================================

/*
==================
=
= StartDemoRecord
=
==================
*/

void FinishDemoRecord (void) {return;}
void RecordDemo (void) {return;}


//==========================================================================

/*
==================
=
= Died
=
==================
*/

#define DEATHROTATE 2

void Died (void)
{
    float   fangle;
    int32_t dx,dy;
    int     iangle,curangle,clockwise,counter,change;
    int     i;

    if (screenfaded)
    {
        ThreeDRefresh ();
        VW_FadeIn ();
    }

    gamestate.weapon = wp_noweapon;                     // take away weapon
    PlaySoundGlobal (PLAYERDEATHSND,chan_weapon_fire,0);

    //
    // swing around to face attacker
    //
    if(killerobj)
    {
        dx = killerobj->x - player->x;
        dy = player->y - killerobj->y;

        fangle = (float) atan2((float) dy, (float) dx);     // returns -pi to pi
        if (fangle<0)
            fangle = (float) (M_PI*2+fangle);

        iangle = (int) (fangle/(M_PI*2)*ANGLES);
    }
    else
    {
        ThreeDRefresh ();
        goto deathfade;
    }

    if (player->angle > iangle)
    {
        counter = player->angle - iangle;
        clockwise = ANGLES-player->angle + iangle;
    }
    else
    {
        clockwise = iangle - player->angle;
        counter = player->angle + ANGLES-iangle;
    }

    curangle = player->angle;

    if (clockwise<counter)
    {
        //
        // rotate clockwise
        //
        if (curangle>iangle)
            curangle -= ANGLES;
        do
        {
            change = tics*DEATHROTATE;
            if (curangle + change > iangle)
                change = iangle-curangle;

            curangle += change;
            player->angle += change;
            if (player->angle >= ANGLES)
                player->angle -= ANGLES;

            ThreeDRefresh ();
            CalcTics ();
        } while (curangle != iangle);
    }
    else
    {
        //
        // rotate counterclockwise
        //
        if (curangle<iangle)
            curangle += ANGLES;
        do
        {
            change = -(int)tics*DEATHROTATE;
            if (curangle + change < iangle)
                change = iangle-curangle;

            curangle += change;
            player->angle += change;
            if (player->angle < 0)
                player->angle += ANGLES;

            ThreeDRefresh ();
            CalcTics ();
        } while (curangle != iangle);
    }

    //
    // fade to red
    //
deathfade:
    FinishPaletteShifts ();

    VL_BarScaledCoord (viewscreenx,viewscreeny,viewwidth,viewheight,4);

    IN_ClearKeysDown ();

    FizzleFade(screenBuffer,screen,viewscreenx,viewscreeny,viewwidth,viewheight,70,false);

    IN_UserInput(100);
    ClearMemory();

#ifdef DEBUGKEYS
    if (!infinitelives)
#endif
        gamestate.lives--;

    if (gamestate.lives > -1)
    {
        for (i=0; i<NUMWEAPONS; i++) { gamestate.hasweapon[i] = 0; }
        for (i=0; i<NUMBUTTONS; i++) { gamestate.lastchosenweapon[i] = (weapontype)0; }
        for (i=0; i<NUMAMMOTYPES; i++) { gamestate.ammo[i] = 0; }
        StartInventory();
        if (gamestate.flakgunparts == NUMFLAKGUNPARTS)
        {
            gamestate.hasweapon[wp_flakgun] = 1;
            gamestate.bestweapon = wp_flakgun;
            gamestate.ammo[ammo_flak] = 100;
        }
        gamestate.keys = 0;
        pwallstate = pwallpos = 0;
        gamestate.attackframe = gamestate.attackcount =
            gamestate.weaponframe = 0;

        if(viewsize != 21)
        {
            DrawKeys ();
            DrawAmmo ();
            DrawHealth ();
            DrawFace ();
            DrawLives ();
        }
    }
}

//==========================================================================

/*
===================
=
= ExitLevel
=
= WSJ: tries to go to the next map
=
===================
*/

void ExitLevel (void)
{
    word tile = MAPSPOT(player->tilex,player->tiley,0);

    switch (tile)    // WSJ: check for secret level tile
    {
        case SECRETEXIT1TILE:
        case SECRETEXIT2TILE:
        case SECRETEXIT3TILE:
            gamestate.nextmap = LRpack+1+tile-SECRETEXIT1TILE;
            break;
    }

    if (gamestate.nextmap >= (LRpack+NUMBONUSMAPS))
        gamestate.nextmap = 0;

    playstate = ex_completed;
}

//==========================================================================

/*
===================
=
= GameLoop
=
===================
*/

void GameLoop (void)
{
    boolean died;
#ifdef MYPROFILE
    clock_t start,end;
#endif

restartgame:
    ClearMemory ();
    SETFONTCOLOR(0,15);
    VW_FadeOut();
    DrawPlayScreen ();
    died = false;
    do
    {
        if (!loadedgame)
            gamestate.score = gamestate.oldscore;
        if(!died || viewsize != 21) DrawScore();

        startgame = false;
        if (!loadedgame)
        {
            SetupGameLevel ();
            if((GetFeatureFlags() & FF_POISON) && gamestate.health < 100)
                gamestate.health = 100;
        }

#ifdef SPEAR
        if (gamestate.mapon == 20)      // give them the key allways
        {
            gamestate.keys |= 1;
            DrawKeys ();
        }
#endif

        ingame = true;
        loadedgame = false;
        StartMusic ();

        if (!died)
            PreloadGraphics ();             // TODO: Let this do something useful!
        else
            died = false;

        fizzlein = true;

        DrawLevel ();

#ifdef SPEAR
startplayloop:
#endif
        PlayLoop ();

#ifdef SPEAR
        if (spearflag)
        {
            SD_StopSound();
            SD_PlaySound(GETSPEARSND);
            Delay(150);

            ClearMemory ();
            gamestate.oldscore = gamestate.score;
            gamestate.mapon = 20;
            SetupGameLevel ();
            StartMusic ();
            player->x = spearx;
            player->y = speary;
            player->angle = (short)spearangle;
            spearflag = false;
            Thrust (0,0);
            goto startplayloop;
        }
#endif

        SD_StopMusic ();
        ingame = false;

        if (demorecord && playstate != ex_warped)
            FinishDemoRecord ();

        if (startgame || loadedgame)
            goto restartgame;

        switch (playstate)
        {
            case ex_completed:
            case ex_secretlevel:
                if(viewsize == 21) DrawPlayScreen();
                gamestate.keys = 0;
                DrawKeys ();
                VW_FadeOut ();
                ClearMemory ();

                LevelCompleted ();              // do the intermission
                if(viewsize == 21) DrawPlayScreen();

                gamestate.oldscore = gamestate.score;

                //
                // GOING TO NEXT LEVEL
                //
                gamestate.mapon = gamestate.nextmap;
#ifdef AUTOINTER
                IntermissionScreens();    // WSJ: INTART screen if file exists
                ClearMemory();            // WSJ: "
                DrawPlayScreen();         // WSJ: "
#endif
                break;

            case ex_died:
                Died ();
                died = true;                    // don't "get psyched!"

                if (gamestate.lives > -1)
                    break;                          // more lives left

                VW_FadeOut ();
                if(screenHeight % 200 != 0)
                    VL_ClearScreen(0);
                ClearMemory ();
                CheckHighScore (gamestate.score,gamestate.mapon+1);
                strcpy(MainMenu[viewscores].string,STR_VS);
                MainMenu[viewscores].routine = CP_ViewScores;
                return;

            case ex_victorious:
                if(viewsize == 21) DrawPlayScreen();
                VL_FadeOut (0,255,0,0,0,200);
                ClearMemory ();
                Victory ();
                ClearMemory ();
                CheckHighScore (gamestate.score,gamestate.mapon+1);
                strcpy(MainMenu[viewscores].string,STR_VS);
                MainMenu[viewscores].routine = CP_ViewScores;
                return;

            default:
                if(viewsize == 21) DrawPlayScreen();
                ClearMemory ();
                break;
        }
    } while (1);
}
